package editor;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.io.*;
import java.text.ParseException;
import java.util.*;

import javax.imageio.ImageIO;
import javax.swing.*;

import chooser.BindingAdaptor;
import chooser.BindingChooser;
import chooser.BindingEntry;
import chooser.Persistence;
import util.ComponentTools;
import util.StringTools;

/**
 * Small graphical application for generating and editing keybindings. This allows them to be stored
 * in various persistent states that can be retrieved later programmatically. This supports all the
 * capabilities of the BindingChooser including ordering, all key types, etc.
 * @author Damian Johnson (atagar1@gmail.com)
 * @version September 2, 2007
 */
public class BindingEditor extends JFrame {
  private static final long serialVersionUID = 0;
  private static final String IMG_BASE_PATH = "images/"; // Relative path of image resources
  private static final String TITLE = "Keybinding Editor";
  private static final Color CONTENT_BACKGROUND = Color.WHITE;
  private static final ImageIcon ACTIVE_ICON; // Arrow to indicate active entry, null if not loaded
  private static final int INITIAL_DISABLING_KEYCODE = KeyEvent.VK_ESCAPE;
  private BindingChooser chooser; // Display of current keybindings
  private JViewport chooserViewport;
  private BindingAdaptor bindingAdaptor; // Key adaptor for editing shortcut field
  private BindingEntry active = null; // None selected when null
  private EditorPersistence persistence;
  private LinkedHashMap <KeyStroke, String> controlBindings;
  private JPanel buttonPanel;
  
  static {
    BufferedImage activeImg;
    try {
      activeImg = getImageResource("active.png");
    } catch (IOException exc) {
      System.out.println(exc.getMessage());
      activeImg = null;
    }
    
    if (activeImg != null) ACTIVE_ICON = new ImageIcon(activeImg);
    else ACTIVE_ICON = null;
  }
  
  {
    /*-
     * Generates custom chooser providing the following capacities:
     * Editing action field
     * Selection of active entry
     */
    this.chooser = new BindingChooser() {
      private static final long serialVersionUID = 0;
      
      protected void onClick(MouseEvent event, BindingEntry entry, BindingEntry.Field field) {
        if (field == BindingEntry.Field.SHORTCUT) {
          super.onClick(event, entry, field); // Uses BindingChooser selection methods
        } else {
          boolean renameAction = field == BindingEntry.Field.ACTION;
          renameAction &= SwingUtilities.isLeftMouseButton(event);
          renameAction &= event.getClickCount() > 1;
          
          if (renameAction) EditorControl.RENAME_ACTION.doAction(BindingEditor.this, entry);
          else {
            if (entry.equals(active)) setActive(null); // Deselects if already active
            else setActive(entry);
          }
        }
      }
      
      protected void onUpdate(int index, BindingEntry entry, boolean isNew) {
        if (entry.equals(active)) {
          applyColorScheme(entry, index); // Prevents chooser from altering indent field
        } else {
          super.onUpdate(index, entry, isNew);
        }
      }
    };
    
    this.bindingAdaptor = this.chooser.makeAdaptor();
    this.bindingAdaptor.setBindingsDisableable(true);
    this.bindingAdaptor.setDisablingKeyCode(INITIAL_DISABLING_KEYCODE);
    
    // Loads editor's keybindings for controls
    String editorBindingsFile = "editorBindings";
    try {
      InputStream stream = BindingEditor.class.getResourceAsStream(editorBindingsFile);
      if (stream == null) throw new IOException();
      this.controlBindings = Persistence.SERIAL_HASH.load(stream);
    } catch (IOException exc) {
      this.controlBindings = new LinkedHashMap <KeyStroke, String> ();
      System.out.println("Unable to load bindings from " + editorBindingsFile);
    } catch (ParseException exc) {
      this.controlBindings = new LinkedHashMap <KeyStroke, String> ();
      System.out.println("Improperly formatted editor bindings:\n" + exc.getMessage());
    }
  }
  
  public static void main(String[] args) {
    BindingEditor editor = new BindingEditor();
    ComponentTools.center(editor);
    editor.setVisible(true);
  }
  
  /**
   * Constructs an empty editor.
   */
  public BindingEditor() {
    this(new HashMap <KeyStroke, String>());
  }
  
  /**
   * Constructs an editor with an initial set of bindings.
   * @param bindings initial contents of the editor
   */
  public BindingEditor(Map <KeyStroke, String> bindings) {
    setTitle(TITLE);
    setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    setResizable(false);
    
    try {
      setIconImage(getImageResource("icon.png"));
    } catch (IOException exc) {
      // Icon is missing- uses default frame icon
    }
    
    this.chooser.putAllBindings(bindings);
    this.persistence = new EditorPersistence(this);
    
    this.buttonPanel = new JPanel(new GridLayout(0, 2)); // Right control panel
    for (EditorButton control : EditorButton.values()) {
      this.buttonPanel.add(control.makeButton(this));
    }
    add(this.buttonPanel, BorderLayout.EAST);
    
    JPanel center = new JPanel(new BorderLayout()); // All content except controls
    JPanel labels = this.chooser.makeLabels(BindingChooser.FrameColorScheme.HUE_DERIVED);
    center.add(labels, BorderLayout.NORTH);
    
    // Allows chooser to keep its preferred size and provides background
    JPanel wrapper = new JPanel(new BorderLayout());
    wrapper.add(this.chooser, BorderLayout.NORTH);
    wrapper.setBackground(CONTENT_BACKGROUND);
    
    int scrollWidth = wrapper.getPreferredSize().width;
    int scrollHeight = this.buttonPanel.getPreferredSize().height;
    JScrollPane scrollPane = new JScrollPane(wrapper);
    scrollPane.setPreferredSize(new Dimension(scrollWidth, scrollHeight));
    scrollPane.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
    scrollPane.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
    this.chooserViewport = scrollPane.getViewport();
    center.add(scrollPane);
    add(center);
    
    /*
     * Listeners for controls and shortcut entry input. Conflicts are avoid by using a listener
     * based on if a shortcut field is selected or not.
     */
    this.addKeyListener(this.bindingAdaptor);
    this.addKeyListener(new KeyAdapter() {
      public void keyPressed(KeyEvent event) {
        if (!chooser.isBindingSelected()) {
          KeyStroke shortcut = KeyStroke.getKeyStrokeForEvent(event);
          if (!controlBindings.containsKey(shortcut)) return;
          String action = controlBindings.get(shortcut);
          
          try {
            EditorButton buttonControl = EditorButton.fromString(action);
            buttonControl.doEvent(BindingEditor.this);
          } catch (IllegalArgumentException exc1) {
            // Action doesn't belong to a button, tries editor controls
            try {
              EditorControl control = EditorControl.fromString(action);
              control.doAction(BindingEditor.this);
            } catch (IllegalArgumentException exc2) {
              // Action doesn't belong to an editor control either- this is an undefined mapping
              assert false : "Unrecognized action: " + action;
            }
          }
        }
      }
    });
    
    pack();
  }
  
  /**
   * Sets the entry to which editor input is directed (moving, removal, etc). The indent field is
   * used to indicate the selection.
   * @param entry target of editor input
   */
  protected void setActive(BindingEntry entry) {
    if (entry != null && entry.equals(this.active)) return; // Entry is already selected
    this.active = entry;
    this.chooser.validate(); // Causes onUpdate call to revert previous active's appearance
    
    // Places arrow in indent field to indicate selection
    if (this.active != null) {
      JLabel indentLabel = this.active.getField(BindingEntry.Field.INDENT);
      
      if (ACTIVE_ICON != null) {
        indentLabel.setText("");
        indentLabel.setIcon(ACTIVE_ICON);
      } else {
        indentLabel.setText(" \u2192"); // Uses Unicode right arrow
        indentLabel.setIcon(null);
      }
      
      setFocus(this.active);
    }
    
    // Sets controls to be enabled or disabled based on if there's an active button
    for (Component component : this.buttonPanel.getComponents()) {
      if (component instanceof JButton) {
        JButton button = (JButton) component;
        try {
          boolean isActiveRequired = EditorButton.fromString(button.getName()).isActiveRequired();
          boolean setEnabled = this.active != null || !isActiveRequired;
          button.setEnabled(setEnabled);
        } catch (IllegalArgumentException exc) {
          button.setEnabled(true); // Button isn't recognized by the EditorButton
        }
      }
    }
  }
  
  /**
   * Provides binding chooser that contains content.
   * @return contained chooser with binding entry content
   */
  protected BindingChooser getChooser() {
    return this.chooser;
  }
  
  /**
   * Provides the keyboard adaptor used to set selected shortcut fields.
   * @return BindingAdaptor used to set shortcut fields
   */
  protected BindingAdaptor getAdaptor() {
    return this.bindingAdaptor;
  }
  
  /**
   * Returns if an entry is awaiting editor input.
   * @return true if there's an entry currently active, false otherwise
   */
  protected boolean isBindingActive() {
    return this.active != null;
  }
  
  /**
   * Returns the entry currently awaiting editor input (null if none is active).
   * @return currently active binding entry
   */
  protected BindingEntry getActiveEntry() {
    return this.active;
  }
  
  // Provides hue on which theme components can be created
  protected Color getReferenceColor() {
    return BindingChooser.FrameColorScheme.HUE_DERIVED.getBackgroundBasis(getChooser());
  }
  
  // Provides module responsible for the persistence of the editor's bindings
  protected EditorPersistence getPersistence() {
    return this.persistence;
  }
  
  /**
   * Attempts to dispose of this editor, prompting user input if the content has changed since last
   * saved. If so, the user is given the option to save, close, or leave the editor open.
   */
  public void dispose() {
    if (this.persistence.isPreviousPathKnown()) {
      if (this.persistence.isUnsaved()) {
        String message = "Bindings have been modified. Save changes?";
        int options = JOptionPane.YES_NO_CANCEL_OPTION;
        int input = JOptionPane.showConfirmDialog(this, message, "Save Changes?", options);
        
        if (input == JOptionPane.CANCEL_OPTION) return;
        else if (input == JOptionPane.YES_OPTION) {
          this.persistence.save();
        }
      }
    } else {
      if (this.persistence.isUnsaved()) {
        String message = "Unsaved changes will be lost. Are you sure you want to quit?";
        int options = JOptionPane.YES_NO_OPTION;
        int input = JOptionPane.showConfirmDialog(this, message, "Save Changes?", options);
        if (input == JOptionPane.NO_OPTION) return;
      }
    }
    
    super.dispose();
  }
  
  /**
   * Displays dialog allowing user to redefine keybindings used to control this editor.
   */
  protected void redefineControlBindings() {
    LinkedHashMap <KeyStroke, String> input = BindingChooser.showDialog(this, this.controlBindings);
    if (input != null) this.controlBindings = input;
  }
  
  // Shifts the view if necessary to make an entry visible
  protected void setFocus(BindingEntry entry) {
    Rectangle visibleArea = this.chooserViewport.getViewRect();
    Point activeLoc = entry.getLocation();
    int activeHeight = entry.getField(BindingEntry.Field.SHORTCUT).getHeight();
    
    if (activeLoc.y < visibleArea.y) {
      // Active is above the visible area
      this.chooserViewport.setViewPosition(entry.getLocation());
    } else if ((activeLoc.y + activeHeight) > (visibleArea.y + visibleArea.height)) {
      // Active is below the visible area
      int newViewHeight = activeLoc.y + activeHeight - visibleArea.height;
      this.chooserViewport.setViewPosition(new Point(0, newViewHeight));
    }
  }
  
  // Gets an image from the image folder
  protected static BufferedImage getImageResource(String resource) throws IOException {
    String path = IMG_BASE_PATH + resource;
    InputStream stream = BindingEditor.class.getResourceAsStream(path);
    if (stream == null) throw new IOException("Image resource not found: " + path);
    BufferedImage img = ImageIO.read(stream);
    stream.close();
    return img;
  }
  
  /**
   * Editor controls provided by this class.
   */
  private enum EditorControl {
    SELECT_UP, SELECT_DOWN, SELECT_TOP, SELECT_BOTTOM, RENAME_ACTION, CLEAR_ENTRIES, QUICK_SAVE;
    
    /**
     * Returns the enum representation of a string. This is case sensitive.
     * @param str toString representation of this enum
     * @return enum associated with a string
     * @throws IllegalArgumentException if argument is not represented by this enum.
     */
    protected static EditorControl fromString(String str) {
      for (EditorControl type : EditorControl.values()) {
        if (str.equals(type.toString())) return type;
      }
      throw new IllegalArgumentException();
    }
    
    /**
     * Fires event associated with this control on the currently active entry. Nothing happens if
     * their isn't an entry currently active but one's required.
     * @param editor binding editor to be effected by the control
     */
    protected void doAction(BindingEditor editor) {
      if (editor.isBindingActive()) doAction(editor, editor.getActiveEntry());
      else doAction(editor, null);
    }
    
    /**
     * Fires event associated with this control on a given binding.
     * @param editor binding editor to be effected by the control
     * @param subject entry on which to perform action
     */
    protected void doAction(BindingEditor editor, BindingEntry subject) {
      BindingChooser chooser = editor.getChooser();
      
      // Actions that don't require a selected entry
      if (this == CLEAR_ENTRIES) {
        editor.getChooser().clearBindings();
      } else if (this == QUICK_SAVE) {
        if (editor.persistence.isPreviousPathKnown()) {
          editor.persistence.save();
        } else {
          JOptionPane.showMessageDialog(editor, "No known path exists to save to.");
        }
      }
      
      // Actions that require a selected entry
      if (subject == null) return;
      
      if (this == SELECT_UP || this == SELECT_DOWN) {
        int index = chooser.getBindingIndex(subject);
        int newIndex =
            this == SELECT_UP ? Math.max(index - 1, 0) : Math.min(index + 1, chooser
                .getBindingCount() - 1);
        editor.setActive(chooser.getBinding(newIndex));
      } else if (this == SELECT_TOP) {
        editor.setActive(chooser.getBinding(0));
      } else if (this == SELECT_BOTTOM) {
        editor.setActive(chooser.getBinding(chooser.getBindingCount() - 1));
      } else if (this == RENAME_ACTION) {
        // Prompts user to rename the action component of a binding
        String message = "Set action to:";
        String input = JOptionPane.showInputDialog(editor, message, subject.getAction());
        if (input != null && !input.equals("")) {
          // If a disabled entry, checks for duplicates and re-prompts if found
          if (subject.getShortcut() == BindingEntry.DISABLED) {
            BindingEntry newEntry = new BindingEntry(BindingEntry.DISABLED, input);
            if (chooser.getBindings().contains(newEntry)) {
              message = "Unable to set action. That input\n";
              message += "would cause a duplicate entry.";
              JOptionPane.showMessageDialog(editor, message, "Duplicate error",
                  JOptionPane.ERROR_MESSAGE);
              return;
            }
          }
          
          subject.setAction(input);
        }
      } else {
        assert false : "No action defined for '" + this + "'";
      }
    }
    
    @Override
    public String toString() {
      if (this == SELECT_UP) return "Select - Up";
      else if (this == SELECT_DOWN) return "Select - Down";
      else if (this == SELECT_TOP) return "Select - Top";
      else if (this == SELECT_BOTTOM) return "Select - Bottom";
      
      return StringTools.getReadableConstant(this.name());
    }
  }
}
