package editor;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.io.*;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Scanner;

import javax.swing.*;

import chooser.BindingAdaptor;
import chooser.BindingChooser;
import util.ComponentTools;
import util.StringTools;

/**
 * Configuration dialog for BindingEditor settings and miscellaneous options.
 * @author Damian Johnson (atagar1@gmail.com)
 * @version August 26, 2007
 */
class ConfigDialog extends JDialog {
  private static final long serialVersionUID = 0;
  private static final String TITLE = "Configurations";
  private static final String ABOUT_TEXT; // Textual content of about dialog
  private static final ImageIcon ABOUT_BANNER; // Top banner of about dialog, null if not loaded
  private static final Map <Integer, String> KEYSTROKE_RELATIONS; // Maps keystrokes to their names
  private static final Dimension CONTROL_DIM = new Dimension(95, 25);
  private static final int BUTTON_HEIGHT = 25;
  private JComboBox eventTypeBox = new JComboBox();
  private JComboBox indentStyleBox = new JComboBox();
  private JCheckBox disablingKeycodeCheckbox;
  private JButton disablingKeycodeSelection = new JButton();
  private int disablingKeycode;
  private BindingEditor editor;
  
  static {
    // Loads banner of about dialog
    BufferedImage bannerImg;
    try {
      bannerImg = BindingEditor.getImageResource("banner.png");
    } catch (IOException exc) {
      System.out.println(exc.getMessage());
      bannerImg = null;
    }
    
    if (bannerImg != null) ABOUT_BANNER = new ImageIcon(bannerImg);
    else ABOUT_BANNER = null;
    
    // Loads text of about dialog
    String path = "about.txt";
    InputStream stream = BindingEditor.class.getResourceAsStream(path);
    if (stream == null) {
      System.out.println("Unable to load about text from '" + path + "'.");
      ABOUT_TEXT = "";
    } else {
      StringBuilder builder = new StringBuilder();
      Scanner scanner = new Scanner(stream);
      
      if (scanner.hasNextLine()) builder.append(scanner.nextLine());
      while (scanner.hasNextLine()) {
        builder.append("\n");
        builder.append(scanner.nextLine());
      }
      ABOUT_TEXT = builder.toString();
      
      try {
        stream.close();
      } catch (IOException exc) {
        // Suppressing warning since stream's finalizer will close resource
      }
    }
    
    KEYSTROKE_RELATIONS = new LinkedHashMap <Integer, String>();
    KEYSTROKE_RELATIONS.put(KeyEvent.KEY_PRESSED, "Pressed");
    KEYSTROKE_RELATIONS.put(KeyEvent.KEY_RELEASED, "Released");
    KEYSTROKE_RELATIONS.put(KeyEvent.KEY_TYPED, "Typed");
  }
  
  {
    this.eventTypeBox.setForeground(Color.BLACK);
    this.eventTypeBox.setFocusable(false);
    this.eventTypeBox.setPreferredSize(CONTROL_DIM);
    for (String keystrokeType : KEYSTROKE_RELATIONS.values()) {
      this.eventTypeBox.addItem(keystrokeType);
    }
    
    this.indentStyleBox.setForeground(Color.BLACK);
    this.indentStyleBox.setFocusable(false);
    this.indentStyleBox.setPreferredSize(CONTROL_DIM);
    this.indentStyleBox.addItem(BindingChooser.IndentStyle.SPACER.toString());
    this.indentStyleBox.addItem(BindingChooser.IndentStyle.TYPE.toString());
    this.indentStyleBox.addItem(BindingChooser.IndentStyle.INDEX.toString());
    
    this.disablingKeycodeCheckbox = new JCheckBox("Disabling Keycode");
    this.disablingKeycodeCheckbox.setForeground(Color.BLACK);
    this.disablingKeycodeCheckbox.setFocusable(false);
    this.disablingKeycodeCheckbox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        boolean isSelected = ((JCheckBox) event.getSource()).isSelected();
        disablingKeycodeSelection.setEnabled(isSelected);
      }
    });
    
    this.disablingKeycodeSelection.setForeground(Color.BLACK);
    this.disablingKeycodeSelection.setFocusable(false);
    this.disablingKeycodeSelection.setPreferredSize(CONTROL_DIM);
    this.disablingKeycodeSelection.setBorder(BorderFactory.createLineBorder(Color.GRAY));
    this.disablingKeycodeSelection.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        final JDialog messageDialog = new JDialog(ConfigDialog.this, true);
        messageDialog.setUndecorated(true);
        messageDialog.addKeyListener(new KeyAdapter() {
          public void keyPressed(KeyEvent event) {
            disablingKeycode = event.getKeyCode();
            disablingKeycodeSelection.setText(KeyEvent.getKeyText(disablingKeycode));
            messageDialog.dispose();
          }
        });
        
        JLabel label = new JLabel("Press any key.");
        label.setForeground(Color.BLACK);
        label.setFont(new Font("Helvetica", Font.BOLD | Font.ITALIC, 25));
        label.setBorder(BorderFactory.createRaisedBevelBorder());
        label.setOpaque(true);
        label.setBackground(new Color(224, 224, 224));
        messageDialog.add(label);
        
        messageDialog.pack();
        ComponentTools.center(messageDialog, ConfigDialog.this);
        messageDialog.setVisible(true);
      }
    });
  }
  
  /**
   * Displays dialog prompting user to change editor configurations.
   * @param editor binding editor generating configuration dialog
   * @return collection of parameters reflecting settings
   */
  protected static Output showDialog(BindingEditor editor) {
    Output initialValues = new Output(editor.getChooser(), editor.getAdaptor());
    ConfigDialog dialog = new ConfigDialog(editor, initialValues);
    
    ComponentTools.center(dialog, editor);
    dialog.setVisible(true);
    return new Output(dialog);
  }
  
  private ConfigDialog(BindingEditor editor, Output initialValues) {
    super(editor, true);
    setTitle(TITLE);
    setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
    setResizable(false);
    
    this.editor = editor;
    Box center = Box.createVerticalBox();
    
    // Adds line for selecting type of keystrokes
    JPanel eventTypeLine = new JPanel(new BorderLayout());
    eventTypeLine.add(makeLabel("Detected Keystrokes"), BorderLayout.WEST);
    this.eventTypeBox.setSelectedItem(KEYSTROKE_RELATIONS.get(initialValues.inputEventType));
    eventTypeLine.add(this.eventTypeBox, BorderLayout.EAST);
    center.add(eventTypeLine);
    
    // Adds line for selecting the style of indent field
    JPanel indentStyleLine = new JPanel(new BorderLayout());
    indentStyleLine.add(makeLabel("Indent Field Content"), BorderLayout.WEST);
    this.indentStyleBox.setSelectedItem(initialValues.indentStyle.toString());
    indentStyleLine.add(this.indentStyleBox, BorderLayout.EAST);
    center.add(indentStyleLine);
    
    // Adds line for setting a disabling keycode
    JPanel disablingKeycodeLine = new JPanel(new BorderLayout());
    this.disablingKeycodeCheckbox.setSelected(initialValues.isEntriesDisableable);
    disablingKeycodeLine.add(this.disablingKeycodeCheckbox, BorderLayout.WEST);
    this.disablingKeycode = initialValues.disablingKeycode;
    this.disablingKeycodeSelection.setText(KeyEvent.getKeyText(initialValues.disablingKeycode));
    this.disablingKeycodeSelection.setEnabled(initialValues.isEntriesDisableable);
    disablingKeycodeLine.add(this.disablingKeycodeSelection, BorderLayout.EAST);
    center.add(disablingKeycodeLine);
    add(center);
    
    JPanel buttonPanel = new JPanel(new GridLayout(0, 2));
    for (Button type : Button.values()) {
      buttonPanel.add(type.makeButton(this));
    }
    add(buttonPanel, BorderLayout.SOUTH);
    pack();
  }
  
  /**
   * Collection of parameters defined by dialog's fields.
   */
  protected static class Output {
    private int inputEventType;
    private BindingChooser.IndentStyle indentStyle;
    private boolean isEntriesDisableable;
    private int disablingKeycode;
    
    private Output(ConfigDialog dialog) {
      // Sets inputEventType to event code represented by eventTypeBox
      String selectedType = (String) dialog.eventTypeBox.getSelectedItem();
      if (KEYSTROKE_RELATIONS.values().contains(selectedType)) {
        for (Integer eventType : KEYSTROKE_RELATIONS.keySet()) {
          String representation = KEYSTROKE_RELATIONS.get(eventType);
          if (representation.equals(selectedType)) {
            this.inputEventType = eventType;
          }
        }
      } else {
        String message = "Unrecognized keyboard event type: " + selectedType;
        throw new IllegalArgumentException(message);
      }
      
      String selectedIndentStyle = (String) dialog.indentStyleBox.getSelectedItem();
      this.indentStyle = BindingChooser.IndentStyle.fromString(selectedIndentStyle);
      this.isEntriesDisableable = dialog.disablingKeycodeCheckbox.isSelected();
      this.disablingKeycode = dialog.disablingKeycode;
    }
    
    protected Output(BindingChooser chooser, BindingAdaptor adaptor) {
      this.inputEventType = adaptor.getInputEventType();
      this.indentStyle = chooser.getIndentStyle();
      this.isEntriesDisableable = adaptor.isBindingDisablingEnabled();
      this.disablingKeycode = adaptor.getDisablingKeyCode();
    }
    
    protected void apply(BindingChooser chooser, BindingAdaptor adaptor) {
      adaptor.setBindingsDisableable(this.isEntriesDisableable);
      adaptor.setDisablingKeyCode(this.disablingKeycode);
      adaptor.setInputEventType(this.inputEventType);
      chooser.setIndentStyle(this.indentStyle);
    }
  }
  
  private static JLabel makeLabel(String text) {
    JLabel label = new JLabel(" " + text);
    label.setForeground(Color.BLACK);
    return label;
  }
  
  // Bottom controls of dialog
  private enum Button {
    CONTROL_BINDINGS, DEMO_CHOOSER, ABOUT, CLOSE;
    
    protected JButton makeButton(final ConfigDialog dialog) {
      Color reference = dialog.editor.getReferenceColor();
      JButton button = ComponentTools.makeThemedButton(toString(), reference);
      button.setFocusable(false); // Prevents interference with setting disabling keycode
      int preferredWidth = button.getPreferredSize().width;
      button.setPreferredSize(new Dimension(preferredWidth, BUTTON_HEIGHT));
      button.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent event) {
          Button.this.doEvent(dialog);
        }
      });
      
      if (this == CLOSE) {
        // Makes dialog close when escape is pressed
        KeyStroke closingKeystroke = KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0);
        button.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(closingKeystroke, "close");
        button.getActionMap().put("close", new AbstractAction() {
          private static final long serialVersionUID = 0;
          
          public void actionPerformed(ActionEvent event) {
            ((JButton) event.getSource()).doClick();
          }
        });
      }
      return button;
    }
    
    protected void doEvent(final ConfigDialog dialog) {
      BindingEditor editor = dialog.editor;
      if (this == CONTROL_BINDINGS) {
        editor.redefineControlBindings();
      } else if (this == DEMO_CHOOSER) {
        Color reference = editor.getReferenceColor();
        DemoDialog.showDialog(editor, editor.getChooser().getBindings(), reference);
      } else if (this == ABOUT) {
        final JDialog aboutDialog = new JDialog(dialog, "About", true);
        aboutDialog.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
        aboutDialog.setResizable(false);
        
        if (ABOUT_BANNER != null) aboutDialog.add(new JLabel(ABOUT_BANNER), BorderLayout.NORTH);
        
        JTextArea textArea = new JTextArea(ABOUT_TEXT);
        textArea.setLineWrap(true);
        textArea.setWrapStyleWord(true);
        textArea.setEditable(false);
        aboutDialog.add(new JScrollPane(textArea));
        
        JButton close = ComponentTools.makeThemedButton("Close", dialog.editor.getReferenceColor());
        int preferredWidth = close.getPreferredSize().width;
        close.setPreferredSize(new Dimension(preferredWidth, BUTTON_HEIGHT));
        close.setFocusable(false);
        ComponentTools.setKeyBinding(close, KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0));
        close.addActionListener(new ActionListener() {
          public void actionPerformed(ActionEvent event) {
            aboutDialog.dispose();
          }
        });
        aboutDialog.add(close, BorderLayout.SOUTH);
        
        aboutDialog.pack();
        ComponentTools.center(aboutDialog, dialog);
        aboutDialog.setVisible(true);
      } else if (this == CLOSE) {
        dialog.dispose();
      }
    }
    
    @Override
    public String toString() {
      if (this == CONTROL_BINDINGS || this == DEMO_CHOOSER || this == ABOUT) {
        return StringTools.getReadableConstant(this.name()) + "...";
      } else {
        return StringTools.getReadableConstant(this.name());
      }
    }
  }
}
