package editor;

import java.awt.*;
import java.awt.event.*;
import java.util.ArrayList;
import java.util.LinkedHashMap;

import javax.swing.*;
import javax.swing.border.TitledBorder;

import chooser.BindingChooser;
import chooser.BindingEntry;

import util.ComponentTools;

/**
 * Settings dialog with options for how binding choosers appear.
 * @author Damian Johnson (atagar1@gmail.com)
 * @version August 26, 2007
 */
class DemoDialog extends JDialog {
  private static final long serialVersionUID = 0;
  private static final String TITLE = "Demo Settings";
  private static final int BUTTON_HEIGHT = 25;
  private static final int CONTROL_GAP = 10; // Minimum distance between controls and labels
  private static final Dimension CONTROL_DIM = new Dimension(135, 23);
  
  /**
   * Constructs a dialog providing the ability to configure and construct binding chooser demos that
   * reflect a set of bindings.
   * @param parent frame to which modal properties are applied
   * @param bindings bindings which which generated binding choosers are populated
   * @param reference reference color from which to generate color scheme for buttons
   */
  protected static void showDialog(Frame parent, ArrayList <BindingEntry> bindings, Color reference) {
    DemoDialog dialog = new DemoDialog(parent, bindings, reference);
    ComponentTools.center(dialog, parent);
    dialog.setVisible(true);
  }
  
  private DemoDialog(final Frame parent, final ArrayList <BindingEntry> bindings,
      Color reference) {
    super(parent, TITLE, true);
    setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
    setResizable(false);
    
    // Adds center area with settings for custom choosers
    final Box center = Box.createVerticalBox();
    center.setBorder(BorderFactory.createTitledBorder("Custom Settings"));
    
    JPanel labelsLine = new JPanel(new BorderLayout());
    final JCheckBox labelsCheckbox = new JCheckBox("Show Labels", true);
    labelsCheckbox.setForeground(Color.BLACK);
    labelsCheckbox.setFocusable(false);
    labelsLine.add(labelsCheckbox, BorderLayout.WEST);
    center.add(labelsLine);
    
    JPanel disabledEntriesLine = new JPanel(new BorderLayout());
    final JCheckBox disabledEntriesCheckbox = new JCheckBox("Show Disabled Entries", false);
    disabledEntriesCheckbox.setForeground(Color.BLACK);
    disabledEntriesCheckbox.setFocusable(false);
    disabledEntriesLine.add(disabledEntriesCheckbox, BorderLayout.WEST);
    center.add(disabledEntriesLine);
    
    JPanel colorSchemeLine = new JPanel(new BorderLayout());
    colorSchemeLine.add(makeLabel("Color Scheme"), BorderLayout.WEST);
    final String LIGHT_STATE = "light"; // HACK: Uses button's name to present state (light or dark)
    final JButton lightenedButton = new JButton();
    lightenedButton.setBorder(BorderFactory.createLineBorder(Color.GRAY));
    lightenedButton.setFocusable(false);
    lightenedButton.addActionListener(new ActionListener() {
      boolean isSelected = false;
      
      public void actionPerformed(ActionEvent event) {
        this.isSelected = !this.isSelected;
        String text = this.isSelected ? "Light" : "Dark";
        String state = this.isSelected ? LIGHT_STATE : "not " + LIGHT_STATE;
        Color background = this.isSelected ? Color.WHITE : Color.BLACK;
        Color foreground = this.isSelected ? Color.BLACK : Color.WHITE;
        
        JButton source = (JButton) event.getSource();
        source.setText(text);
        source.setName(state);
        source.setBackground(background);
        source.setForeground(foreground);
      }
    });
    lightenedButton.doClick(); // Initialized background and text
    
    final JButton hueButton = new JButton("Hue");
    hueButton.setBorder(BorderFactory.createLineBorder(Color.GRAY));
    hueButton.setFocusable(false);
    hueButton.setForeground(Color.BLACK);
    hueButton.setBackground(new Color(110, 110, 255));
    hueButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        Color selected = JColorChooser.showDialog(parent, "Hue...", hueButton.getBackground());
        if (selected != null) {
          JButton source = (JButton) event.getSource();
          source.setBackground(selected);
          
          int rgbSum = selected.getRed() + selected.getGreen() + selected.getBlue();
          Color foreground = rgbSum < 128 ? Color.WHITE : Color.BLACK;
          source.setForeground(foreground);
        }
      }
    });
    
    JPanel colorButtons = new JPanel(new GridLayout(0, 2));
    colorButtons.setPreferredSize(CONTROL_DIM);
    colorButtons.add(lightenedButton);
    colorButtons.add(hueButton);
    colorSchemeLine.add(colorButtons, BorderLayout.EAST);
    center.add(colorSchemeLine);
    
    JPanel indentStyleLine = new JPanel(new BorderLayout());
    indentStyleLine.add(makeLabel("Indent Style"), BorderLayout.WEST);
    final JComboBox indentStyleBox = new JComboBox();
    indentStyleBox.setForeground(Color.BLACK);
    indentStyleBox.setFocusable(false);
    indentStyleBox.setPreferredSize(CONTROL_DIM);
    for (BindingChooser.IndentStyle style : BindingChooser.IndentStyle.values()) {
      if (style == BindingChooser.IndentStyle.NONE) continue;
      indentStyleBox.addItem(style.toString());
    }
    indentStyleBox.setSelectedItem(BindingChooser.IndentStyle.INDEX.toString());
    indentStyleLine.add(indentStyleBox, BorderLayout.EAST);
    center.add(indentStyleLine);
    
    JPanel frameSchemeLine = new JPanel(new BorderLayout());
    frameSchemeLine.add(makeLabel("Frame Color"), BorderLayout.WEST);
    final JComboBox frameSchemeBox = new JComboBox();
    frameSchemeBox.setForeground(Color.BLACK);
    frameSchemeBox.setFocusable(false);
    frameSchemeBox.setPreferredSize(CONTROL_DIM);
    for (BindingChooser.FrameColorScheme scheme : BindingChooser.FrameColorScheme.values()) {
      frameSchemeBox.addItem(scheme.toString());
    }
    frameSchemeBox.setSelectedItem(BindingChooser.FrameColorScheme.HUE_DERIVED.toString());
    frameSchemeLine.add(frameSchemeBox, BorderLayout.EAST);
    center.add(frameSchemeLine);
    add(center);
    
    // Adds top option for default appearance
    final JCheckBox defaultCheckbox = new JCheckBox("Default Settings");
    defaultCheckbox.setForeground(Color.BLACK);
    defaultCheckbox.setFocusable(false);
    defaultCheckbox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        boolean isCustom = !((JCheckBox) event.getSource()).isSelected();
        Color titleColor = isCustom ? Color.BLACK : Color.GRAY;
        ((TitledBorder) center.getBorder()).setTitleColor(titleColor);
        setComponentsEnabled(center, isCustom);
      }
      
      // Recursively sets all components to be enabled or disabled
      private void setComponentsEnabled(Component comp, boolean enabled) {
        if (comp instanceof Container) {
          for (Component subcomp : ((Container) comp).getComponents()) {
            setComponentsEnabled(subcomp, enabled);
          }
        }
        
        comp.setEnabled(enabled);
      }
    });
    add(defaultCheckbox, BorderLayout.NORTH);
    defaultCheckbox.doClick(); // Disables custom fields
    
    // Adds bottom controls
    JPanel bottom = new JPanel(new GridLayout(1, 0));
    bottom.setPreferredSize(new Dimension(bottom.getPreferredSize().width, BUTTON_HEIGHT));
    JButton show = ComponentTools.makeThemedButton("Show...", reference);
    show.setFocusable(false);
    show.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        if (defaultCheckbox.isSelected()) {
          LinkedHashMap <KeyStroke, String> bindingMap = new LinkedHashMap <KeyStroke, String> ();
          for (BindingEntry entry : bindings) {
            if (entry.getShortcut() != BindingEntry.DISABLED) {
              bindingMap.put(entry.getShortcut(), entry.getAction());
            }
          }
          
          BindingChooser.showDialog(parent, bindingMap);
        } else {
          boolean showLabels = labelsCheckbox.isSelected();
          boolean showDisabledEntries = disabledEntriesCheckbox.isSelected();
          boolean lightened = lightenedButton.getName().equals(LIGHT_STATE);
          Color hue = hueButton.getBackground();
          
          String indentSelection = (String) indentStyleBox.getSelectedItem();
          BindingChooser.IndentStyle indentStyle =
              BindingChooser.IndentStyle.fromString(indentSelection);
          
          String frameSelection = (String) frameSchemeBox.getSelectedItem();
          BindingChooser.FrameColorScheme frameScheme =
              BindingChooser.FrameColorScheme.fromString(frameSelection);
          
          BindingChooser chooser = new BindingChooser();
          for (BindingEntry entry : bindings) {
            if (showDisabledEntries || entry.getShortcut() != BindingEntry.DISABLED) {
              chooser.putBinding(entry.getShortcut(), entry.getAction());
            }
          }
          chooser.setIndentStyle(indentStyle);
          chooser.setColorScheme(hue, lightened);
          BindingChooser.showDialog(parent, chooser, "Demo Chooser", showLabels, chooser
              .makeAdaptor(), frameScheme);
        }
      }
    });
    bottom.add(show);
    
    JButton close = ComponentTools.makeThemedButton("Close", reference);
    close.setFocusable(false);
    ComponentTools.setKeyBinding(close, KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0));
    close.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        dispose();
      }
    });
    
    bottom.add(close);
    add(bottom, BorderLayout.SOUTH);
    
    // Adds gap to separate controls from labels
    Dimension dialogDim = getPreferredSize();
    setPreferredSize(new Dimension(dialogDim.width + CONTROL_GAP, dialogDim.height));
    pack();
  }
  
  private static JLabel makeLabel(String text) {
    JLabel label = new JLabel(" " + text);
    label.setForeground(Color.BLACK);
    return label;
  }
}
