package editor;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.HashSet;

import javax.swing.*;

import chooser.BindingChooser;
import chooser.BindingEntry;
import util.ComponentTools;
import util.StringTools;

/**
 * Button controls available in the BindingEditor. This provides both a simple means of generating
 * UI components and calling the associated actions programmatically. Actions are as follows:<br>
 * MOVE_UP- Moves active entry up an index.<br>
 * MOVE_DOWN- Moves active entry down an index.<br>
 * MOVE_TOP- Moves active entry to the top of the listing.<br>
 * MOVE_BOTTOM- Moves active entry to the bottom of the listing.<br>
 * ADD_ENTRY- Adds a blank mapping to the end of the listing.<br>
 * REMOVE_ENTRY- Removes the currently active entry and selects next one from the listing.<br>
 * LOAD- Displays dialog with options for retrieving bindings.<br>
 * SAVE- Displays dialog with options for saving current bindings to a file.<br>
 * OPTIONS- Displays configuration dialog.<br>
 * CLOSE- Closes editor, prompting confirmation if contents have changed.
 * @author Damian Johnson (atagar1@gmail.com)
 * @version August 7, 2007
 */
enum EditorButton {
  MOVE_UP("up.png", true), MOVE_DOWN("down.png", true), MOVE_TOP("top.png", true), MOVE_BOTTOM(
      "bottom.png", true), ADD_ENTRY("add.png", false), REMOVE_ENTRY("remove.png", true), LOAD(
      "load.png", false), SAVE("save.png", false), CONFIG("config.png", false), CLOSE(
      "close.png", false);
  
  private final ImageIcon buttonIcon; // Cached button icon
  private final boolean requiresActive; // If button requires active field to take action
  private static final String NEW_ENTRY_ACTION = "Undefined";
  private static final int BUTTON_WIDTH = 48;
  private static final int BUTTON_HEIGHT = 48;
  
  private EditorButton(String iconName, boolean requiresActive) {
    this.requiresActive = requiresActive;
    
    BufferedImage img = new BufferedImage(BUTTON_WIDTH, BUTTON_HEIGHT, BufferedImage.TYPE_INT_ARGB);
    try {
      BufferedImage loadedImg = BindingEditor.getImageResource(iconName);
      
      // Scales loaded image to fill button
      double xScale = (double) BUTTON_WIDTH / loadedImg.getWidth();
      double yScale = (double) BUTTON_HEIGHT / loadedImg.getHeight();
      Graphics2D g2 = (Graphics2D) img.getGraphics();
      g2.scale(xScale, yScale);
      g2.drawImage(loadedImg, 0, 0, null);
    } catch (IOException exc) {
      System.out.println(exc.getMessage());
    }
    this.buttonIcon = new ImageIcon(img);
  }
  
  /**
   * Returns the enum representation of a string. This is case sensitive.
   * @param str toString representation of this enum
   * @return enum associated with a string
   * @throws IllegalArgumentException if argument is not represented by this enum.
   */
  protected static EditorButton fromString(String str) {
    for (EditorButton type : EditorButton.values()) {
      if (str.equals(type.toString())) return type;
    }
    throw new IllegalArgumentException();
  }
  
  /**
   * Returns if control requires an active entry for its listener to take effect.
   * @return true if actions require an active entry, false otherwise
   */
  protected boolean isActiveRequired() {
    return this.requiresActive;
  }
  
  /**
   * Provides button associated with this control, initializing both the appearance and listeners.
   * The name of the component is set to the generating enum's toString representation.
   * @param editor binding editor to be effected by the control
   * @return button with appearance and listeners associated with a specific control
   */
  protected JButton makeButton(final BindingEditor editor) {
    Color reference =
        BindingChooser.FrameColorScheme.HUE_DERIVED.getBackgroundBasis(editor.getChooser());
    JButton button = ComponentTools.makeThemedButton("", reference);
    button.setIcon(this.buttonIcon);
    button.setFocusable(false); // Prevents interference with keyboard listeners
    button.setName(toString()); // Provides means of figuring out what enum generated it
    button.setEnabled(editor.isBindingActive() || !this.requiresActive);
    button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        doEvent(editor);
      }
    });
    
    // Needed for proper layout
    int width = this.buttonIcon.getIconWidth();
    int height = this.buttonIcon.getIconHeight();
    button.setPreferredSize(new Dimension(width, height));
    
    return button;
  }
  
  /**
   * Fires event associated with this control.
   * @param editor binding editor to be effected by the control
   */
  protected void doEvent(BindingEditor editor) {
    if (this.requiresActive && !editor.isBindingActive()) return;
    BindingEntry active = editor.getActiveEntry();
    
    if (this == MOVE_UP || this == MOVE_DOWN || this == MOVE_TOP || this == MOVE_BOTTOM) {
      boolean isUp = this == MOVE_UP || this == MOVE_TOP;
      boolean isEnd = this == MOVE_TOP || this == MOVE_BOTTOM;
      BindingChooser chooser = editor.getChooser();
      
      int newIndex;
      if (!isEnd) {
        int currentIndex = chooser.getBindingIndex(active);
        if (isUp) newIndex = Math.max(currentIndex - 1, 0);
        else newIndex = Math.min(currentIndex + 1, chooser.getBindingCount() - 1);
      } else {
        if (isUp) newIndex = 0;
        else newIndex = chooser.getBindingCount() - 1;
      }
      
      // Removes and re-adds to move entry
      chooser.removeBinding(active);
      chooser.putBinding(active, newIndex);
      editor.validate(); // Ensures change in index is detected before shifting viewport
      editor.setFocus(active);
    } else if (this == ADD_ENTRY) {
      editor.getChooser().putBinding(BindingEntry.DISABLED, getNewEntryAction(editor.getChooser()));
    } else if (this == REMOVE_ENTRY) {
      BindingChooser chooser = editor.getChooser();
      int nextIndex = Math.min(chooser.getBindingIndex(active), chooser.getBindingCount() - 2);
      chooser.removeBinding(active);
      
      if (chooser.getBindingCount() == 0) editor.setActive(null); // Deselects if none's left
      else editor.setActive(chooser.getBinding(nextIndex)); // Selects next entry
      editor.validate(); // Ensures removal of the last entry are detected
    } else if (this == LOAD) {
      editor.getPersistence().showLoadDialog();
    } else if (this == SAVE) {
      editor.getPersistence().showSaveDialog();
    } else if (this == CONFIG) {
      ConfigDialog.Output output = ConfigDialog.showDialog(editor);
      output.apply(editor.getChooser(), editor.getAdaptor());
    } else if (this == CLOSE) {
      editor.dispose();
    }
  }
  
  // Provides unique action by appending a number to the action
  private static String getNewEntryAction(BindingChooser chooser) {
    HashSet <String> actions = new HashSet <String> ();
    for (BindingEntry entry : chooser.getBindings()) {
      actions.add(entry.getAction());
    }
    
    int maxTries = 10000;
    for (int i=1; i < maxTries; ++i) {
      String possibleAction = NEW_ENTRY_ACTION + " " + i;
      if (actions.contains(possibleAction)) continue;
      else return possibleAction;
    }
    
    String message = "Failed to find unique action after " + maxTries + " tries.";
    throw new Error(message);
  }
  
  @Override
  public String toString() {
    if (this == MOVE_UP) return "Move - Up";
    else if (this == MOVE_DOWN) return "Move - Down";
    else if (this == MOVE_TOP) return "Move - Top";
    else if (this == MOVE_BOTTOM) return "Move - Bottom";
    else if (this == REMOVE_ENTRY) return "Remove Selected Entry";
    else return StringTools.getReadableConstant(this.name());
  }
}
