package editor;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.text.ParseException;
import java.util.*;

import javax.swing.*;
import javax.swing.border.TitledBorder;

import chooser.BindingEntry;
import chooser.Persistence;
import util.ComponentTools;

/**
 * Generates dialogs for saving and loading the current state of the editor.
 * @author Damian Johnson (atagar1@gmail.com)
 * @version September 23, 2007
 */
class EditorPersistence {
  private static final int DIALOG_WIDTH = 250; // Width of load/save dialogs
  private static final int BUTTON_HEIGHT = 25; // Height of dialog controls
  private static final Dimension OUTPUT_SIZE = new Dimension(500, 300); // Save output dialog size
  private JFileChooser fileChooser = new JFileChooser();
  private BindingEditor editor;
  private String baseEditorTitle; // Title of editor before appending paths of content
  private ArrayList <BindingEntry> lastSaved; // Used to detect changes in content
  
  /*
   * Provides the last path from which content came (either being loaded or saved) and its
   * associated format. They're null if no IO has taken place yet.
   */
  private String lastUsedPath = null;
  private Persistence lastUsedType = null;
  
  EditorPersistence(BindingEditor editor) {
    this.editor = editor;
    this.baseEditorTitle = this.editor.getTitle();
    this.lastSaved = this.editor.getChooser().getBindings();
  }
  
  /**
   * Returns if the content has changed from the last known preserved state (either having been
   * saved or the content loaded in its entirety). The initially preserved state is considered to be
   * the initial content of the editor.
   * @return true if content has changed since the last known preserved state, false otherwise
   */
  protected boolean isUnsaved() {
    return !this.editor.getChooser().getBindings().equals(this.lastSaved);
  }
  
  /**
   * Returns if previous IO activity has occurred, and hence if the quickSave method is available.
   * @return true if previous IO activity has occurred, false otherwise
   */
  protected boolean isPreviousPathKnown() {
    return this.lastUsedPath != null && this.lastUsedType != null;
  }
  
  /**
   * Saves the editor's current content to the location last saved to or loaded from in the same
   * format as that content.
   * @return true if content was saved, false otherwise (either being canceled or an error
   *         occurring)
   * @throws UnsupportedOperationException if there hasn't been previous IO activity
   */
  protected boolean save() {
    if (!isPreviousPathKnown()) {
      throw new UnsupportedOperationException("No known path or format to save to.");
    } else {
      return save(this.lastUsedType, this.lastUsedPath);
    }
  }
  
  /**
   * Saves the editor's current content, displaying dialogs to mention the success or failure.
   * @param type type of persistence to use in saving bindings
   * @param path location to which bindings are saved
   * @return true if content was saved, false otherwise (either being canceled or an error
   *         occurring)
   */
  protected boolean save(Persistence type, String path) {
    try {
      if (!checkDisabledEntries()) return false;
      type.save(path, this.editor.getChooser().getBindingMap());
      setLastAccessed(path, type);
      JOptionPane.showMessageDialog(editor, "Bindings saved to:\n" + path);
      return true;
    } catch (IOException exc) {
      String message = "Unable to save bindings";
      JOptionPane.showMessageDialog(editor, message, "Save Error", JOptionPane.ERROR_MESSAGE);
      return false;
    }
  }
  
  protected void showLoadDialog() {
    final JDialog loadDialog = new JDialog(this.editor, "Loading Options", true);
    loadDialog.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
    loadDialog.setResizable(false);
    
    Box center = Box.createVerticalBox();
    JPanel typePanel = new JPanel(new GridLayout(0, 1));
    typePanel.setBorder(BorderFactory.createTitledBorder("Format"));
    ((TitledBorder) typePanel.getBorder()).setTitleColor(Color.BLACK);
    final ButtonGroup typeGroup = new ButtonGroup();
    for (Persistence persistenceType : Persistence.values()) {
      boolean isDefaultSelection = persistenceType == Persistence.PROPERTIES_XML;
      typePanel.add(makeRadioButton(persistenceType.toString(), typeGroup, isDefaultSelection));
    }
    center.add(typePanel);
    
    JPanel replacementPanel = new JPanel(new GridLayout(1, 0));
    replacementPanel.setBorder(BorderFactory.createTitledBorder("Current Content"));
    ((TitledBorder) replacementPanel.getBorder()).setTitleColor(Color.BLACK);
    final ButtonGroup replacementGroup = new ButtonGroup();
    final String appendOption = "Append";
    replacementPanel.add(makeRadioButton("Clear", replacementGroup, true));
    replacementPanel.add(makeRadioButton(appendOption, replacementGroup));
    center.add(replacementPanel);
    loadDialog.add(center);
    
    // Adds bottom controls
    JPanel bottom = new JPanel(new GridLayout(1, 0));
    bottom.setPreferredSize(new Dimension(bottom.getPreferredSize().width, BUTTON_HEIGHT));
    JButton load = ComponentTools.makeThemedButton("Load...", this.editor.getReferenceColor());
    load.setFocusable(false);
    load.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        String actionCommand = typeGroup.getSelection().getActionCommand();
        
        try {
          Persistence type = Persistence.fromString(actionCommand);
          int result = fileChooser.showOpenDialog(editor);
          if (result == JFileChooser.APPROVE_OPTION) {
            String loadPath = fileChooser.getSelectedFile().getAbsolutePath();
            try {
              LinkedHashMap <KeyStroke, String> bindings = type.load(loadPath);
              boolean isAppending =
                  replacementGroup.getSelection().getActionCommand().equals(appendOption);
              if (isAppending) {
                // Generates list of replacements that would be made due to duplicate shortcuts
                LinkedHashMap <String, String> replacements = new LinkedHashMap <String, String>();
                HashMap <KeyStroke, String> previousBindings = editor.getChooser().getBindingMap();
                for (KeyStroke shortcut : previousBindings.keySet()) {
                  if (bindings.containsKey(shortcut)) {
                    String previousAction = previousBindings.get(shortcut);
                    String newAction = bindings.get(shortcut);
                    if (previousAction.equals(newAction)) continue;
                    else replacements.put(previousAction, newAction);
                  }
                }
                
                // Confirms that replacements should be made if there are duplicate shortcuts
                if (!replacements.isEmpty()) {
                  StringBuilder message = new StringBuilder();
                  message
                      .append("Duplicate shortcuts exists. The following actions will be replaced:\n");
                  int counter = 1;
                  for (String previousAction : replacements.keySet()) {
                    // Lines are formatted as: (Counter). (Previous Action) -> (New Action)
                    message.append(counter);
                    message.append(". ");
                    message.append(previousAction);
                    message.append(" \u2192 "); // Unicode arrow pointing right
                    message.append(replacements.get(previousAction)); // New action
                    message.append("\n");
                    ++counter;
                  }
                  message.append("Are you sure you want to make these replacements?");
                  
                  int selection =
                      JOptionPane.showConfirmDialog(editor, message.toString(), "Load Error",
                          JOptionPane.YES_NO_OPTION);
                  if (selection == JOptionPane.NO_OPTION) return;
                }
                editor.getChooser().putAllBindings(bindings);
              } else {
                editor.getChooser().clearBindings();
                editor.getChooser().putAllBindings(bindings);
              }
              
              setLastAccessed(loadPath, type);
            } catch (IOException exc) {
              String message = "Unable to load bindings.";
              JOptionPane.showMessageDialog(editor, message, "Load Error",
                  JOptionPane.ERROR_MESSAGE);
            } catch (ParseException exc) {
              String message = "Unable to parse bindings:\n" + exc.getMessage();
              JOptionPane.showMessageDialog(editor, message, "Load Error",
                  JOptionPane.ERROR_MESSAGE);
            }
          }
        } catch (IllegalArgumentException exc) {
          assert false : "Unrecognized type: " + actionCommand;
        }
      }
    });
    bottom.add(load);
    
    JButton close = ComponentTools.makeThemedButton("Close", this.editor.getReferenceColor());
    close.setFocusable(false);
    ComponentTools.setKeyBinding(close, KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0));
    close.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        loadDialog.dispose();
      }
    });
    bottom.add(close);
    loadDialog.add(bottom, BorderLayout.SOUTH);
    
    loadDialog.setPreferredSize(new Dimension(DIALOG_WIDTH, loadDialog.getPreferredSize().height));
    loadDialog.pack();
    ComponentTools.center(loadDialog, this.editor);
    loadDialog.setVisible(true);
  }
  
  /**
   * Displays a modal dialog with saving options.
   */
  protected void showSaveDialog() {
    final JDialog saveDialog = new JDialog(this.editor, "Saving Options", true);
    saveDialog.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
    saveDialog.setResizable(false);
    final ButtonGroup typeGroup = new ButtonGroup();
    final String sourceOption = "Source Code"; // Special option for displaying source code output
    
    // Adds bottom controls
    JPanel bottom = new JPanel(new GridLayout(2, 0));
    bottom.setPreferredSize(new Dimension(bottom.getPreferredSize().width, BUTTON_HEIGHT * 2));
    
    final JButton save = ComponentTools.makeThemedButton("Save", this.editor.getReferenceColor());
    save.setFocusable(false);
    save.setEnabled(isPreviousPathKnown());
    save.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        assert isPreviousPathKnown() : "Attempting to save without knowing the path";
        String actionCommand = typeGroup.getSelection().getActionCommand();
        try {
          save(Persistence.fromString(actionCommand), lastUsedPath);
        } catch (IllegalArgumentException exc) {
          assert false : "Unrecognized type: " + actionCommand;
        }
      }
    });
    bottom.add(save);
    
    final JButton saveAs =
        ComponentTools.makeThemedButton("Save As...", this.editor.getReferenceColor());
    saveAs.setFocusable(false);
    saveAs.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        int result = fileChooser.showSaveDialog(new JDialog());
        if (result == JFileChooser.APPROVE_OPTION) {
          File selected = fileChooser.getSelectedFile();
          
          if (selected.exists()) {
            String message = "File exists. Overwrite?\n" + selected.getAbsolutePath();
            int dialog =
                JOptionPane.showConfirmDialog(null, message, "Confirmation",
                    JOptionPane.YES_NO_OPTION);
            if (dialog == JOptionPane.NO_OPTION) return;
          }
          
          String actionCommand = typeGroup.getSelection().getActionCommand();
          try {
            save(Persistence.fromString(actionCommand), selected.getAbsolutePath());
            save.setEnabled(isPreviousPathKnown());
          } catch (IllegalArgumentException exc) {
            assert false : "Unrecognized type: " + actionCommand;
          }
        }
      }
    });
    bottom.add(saveAs);
    
    JButton showOutput =
        ComponentTools.makeThemedButton("Show Output...", this.editor.getReferenceColor());
    showOutput.setFocusable(false);
    showOutput.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        final JDialog outputDialog = new JDialog(editor, "Output", true);
        outputDialog.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
        outputDialog.setResizable(false);
        outputDialog.setPreferredSize(OUTPUT_SIZE);
        
        String output = "";
        String actionCommand = typeGroup.getSelection().getActionCommand();
        if (actionCommand.equals(sourceOption)) {
          output = getSource(editor.getChooser().getBindingMap());
        } else {
          try {
            Persistence type = Persistence.fromString(actionCommand);
            output = type.getOutput(editor.getChooser().getBindingMap());
          } catch (IllegalArgumentException exc) {
            assert false : "Unrecognized type: " + actionCommand;
          }
        }
        
        JTextArea textArea = new JTextArea(output);
        textArea.setEditable(false);
        outputDialog.add(new JScrollPane(textArea));
        
        JButton close = ComponentTools.makeThemedButton("Close", editor.getReferenceColor());
        int preferredWidth = close.getPreferredSize().width;
        close.setPreferredSize(new Dimension(preferredWidth, BUTTON_HEIGHT));
        close.setFocusable(false);
        ComponentTools.setKeyBinding(close, KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0));
        close.addActionListener(new ActionListener() {
          public void actionPerformed(ActionEvent event) {
            outputDialog.dispose();
          }
        });
        outputDialog.add(close, BorderLayout.SOUTH);
        
        outputDialog.pack();
        ComponentTools.center(outputDialog, editor);
        outputDialog.setVisible(true);
      }
    });
    bottom.add(showOutput);
    
    JButton close = ComponentTools.makeThemedButton("Close", this.editor.getReferenceColor());
    close.setFocusable(false);
    ComponentTools.setKeyBinding(close, KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0));
    close.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        saveDialog.dispose();
      }
    });
    bottom.add(close);
    saveDialog.add(bottom, BorderLayout.SOUTH);
    
    // Adds center area
    JPanel typePanel = new JPanel(new GridLayout(0, 1));
    typePanel.setBorder(BorderFactory.createTitledBorder("Format"));
    ((TitledBorder) typePanel.getBorder()).setTitleColor(Color.BLACK);
    
    // Generates list of options (persistence types plus source)
    ArrayList <String> options = new ArrayList <String> ();
    options.add(sourceOption);
    for (Persistence persistenceType : Persistence.values()) {
      options.add(persistenceType.toString());
    }
    
    for (final String option : options) {
      boolean isDefaultSelection = option.equals(Persistence.PROPERTIES_XML.toString());
      JRadioButton button = makeRadioButton(option, typeGroup, isDefaultSelection);
      button.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent event) {
          boolean isSourceType = option.equals(sourceOption);
          save.setEnabled(!isSourceType && isPreviousPathKnown());
          saveAs.setEnabled(!isSourceType);
        }
      });
      
      typePanel.add(button);
    }
    saveDialog.add(typePanel);
    
    saveDialog.setPreferredSize(new Dimension(DIALOG_WIDTH, saveDialog.getPreferredSize().height));
    saveDialog.pack();
    ComponentTools.center(saveDialog, this.editor);
    saveDialog.setVisible(true);
  }
  
  /**
   * Checks the current content for disabled entries and, if found, confirms that user still wants
   * to save without them.
   * @return true if content should still be saved, false otherwise
   */
  private boolean checkDisabledEntries() {
    for (BindingEntry entry : this.editor.getChooser().getBindings()) {
      if (entry.isDisabled()) {
        String message = "Disabled entries can't be saved. Continue?";
        int input =
            JOptionPane.showConfirmDialog(this.editor, message, "Content Issue",
                JOptionPane.YES_NO_OPTION);
        return input == JOptionPane.YES_OPTION;
      }
    }
    return true; // No disabled entries exist
  }
  
  // Updates reference to where the last content came (either being loaded or saved)
  private void setLastAccessed(String path, Persistence type) {
    this.lastSaved = this.editor.getChooser().getBindings();
    this.lastUsedPath = path;
    this.lastUsedType = type;
    
    // Sets title of editor to reflect the last accessed content
    String title = path == null ? this.baseEditorTitle : this.baseEditorTitle + ": " + path;
    this.editor.setTitle(title);
  }
  
  // Provides a radio button belonging to a specific group. Action command is set to the name.
  private static JRadioButton makeRadioButton(String name, ButtonGroup group) {
    return makeRadioButton(name, group, false);
  }
  
  private static JRadioButton makeRadioButton(String name, ButtonGroup group, boolean isSelected) {
    JRadioButton button = new JRadioButton(name, isSelected);
    button.setActionCommand(name);
    button.setFocusable(false);
    button.setForeground(Color.BLACK);
    group.add(button);
    return button;
  }
  
  /**
   * Generates source code that can be copied directly to generate bindings (disabled entries aren't
   * supported). This reflects: <code>
   * HashMap <KeyStroke, String> keyMap = new HashMap <KeyStroke, String> ();
   * or InputMap keyMap = new InputMap();
   * 
   * for (BindingEntry entry : editor.getChooser().getBindings()) {
   *   keyMap.put(KeyStroke.getKeyStroke(entry.getShortcut().toString()), entry.getAction());
   * }
   * </code>
   * @param bindings keybindings to be constructed by the source code
   * @return string representation of source code able to generate bindings
   */
  private String getSource(Map <KeyStroke, String> bindings) {
    StringBuilder builder = new StringBuilder();
    builder.append("Either:\n");
    builder
        .append("    HashMap <KeyStroke, String> keyMap = new HashMap <KeyStroke, String> ();\n");
    builder.append("    InputMap keyMap = new InputMap();\n");
    builder.append("\n");
    for (KeyStroke shortcut : bindings.keySet()) {
      builder.append("keyMap.put(KeyStroke.getKeyStroke(\"");
      builder.append(shortcut.toString());
      builder.append("\"), \"");
      builder.append(bindings.get(shortcut));
      builder.append("\");\n");
    }
    return builder.toString();
  }
}
