package test.chooser;

import java.awt.*;
import java.awt.event.KeyEvent;
import java.util.ArrayList;
import java.util.LinkedHashMap;

import javax.swing.*;

import chooser.BindingEntry;
import junit.framework.TestCase;

/**
 * Tests for the basic components of the key binding display, exercising both the BindingPanel and
 * BindingEntry classes. This includes a visual component that can be enabled to show the layouts of
 * certain tests.
 * @author Damian Johnson (atagar1@gmail.com)
 * @version June 25, 2007
 */
public class BindingTest extends TestCase {
  /*
   * Milliseconds that test layouts are shown for testClick methods. This includes a description of
   * what layout should be presented. Disabled if duration is less than one. Note that display may
   * be disrupted if tests fail in which its used.
   */
  private static final int LAYOUT_DEMO_DURATION = 0; // Time for substantial change
  private static final int CLICK_DURATION = 0; // Time to indicate a clicked position
  private static final JFrame DEMO_FRAME = new JFrame(); // Frame used to show visual components
  private BindingTestPanel testPanel;
  
  static {
    DEMO_FRAME.setTitle("Binding test demo");
    DEMO_FRAME.setResizable(false);
    
    // Prints explanation of field abbreviations
    if (LAYOUT_DEMO_DURATION > 0) {
      System.out.println("Field key:");
      BindingEntry.Field[] fields = BindingEntry.Field.values();
      for (int i = 0; i < fields.length; ++i) {
        System.out.println("\t" + fields[i].toString().substring(0, 1) + " = " + fields[i]);
      }
      System.out.println("\t*Note: Clicked coordinates are marked by red dots");
    }
  }
  
  public static void main(String args[]) {
    junit.textui.TestRunner.run(BindingTest.class);
  }
  
  // Populates with common key bindings
  public void setUp() {
    this.testPanel = new BindingTestPanel();
    this.testPanel.setCheckAssertions(false);
    
    // Function controls
    this.testPanel.putBinding('L', KeyEvent.CTRL_MASK, "Load");
    this.testPanel.putBinding('S', KeyEvent.CTRL_MASK, "Save");
    
    // Operating system bindings
    this.testPanel.putBinding(KeyEvent.VK_TAB, KeyEvent.ALT_MASK, "Next Application");
    this.testPanel.putBinding(KeyEvent.VK_F4, KeyEvent.ALT_MASK, "Close");
    this.testPanel.putBinding(KeyEvent.VK_DELETE, KeyEvent.CTRL_MASK | KeyEvent.ALT_MASK, "Kill");
    
    // Game controls
    this.testPanel.putBinding(KeyEvent.VK_UP, 0, "Move Up");
    this.testPanel.putBinding(KeyEvent.VK_RIGHT, 0, "Move Right");
    this.testPanel.putBinding(KeyEvent.VK_DOWN, 0, "Move Down");
    this.testPanel.putBinding(KeyEvent.VK_LEFT, 0, "Move Left");
    this.testPanel.putBinding(KeyEvent.VK_CONTROL, 0, "Next Weapon");
    this.testPanel.putBinding(KeyEvent.VK_SPACE, 0, "Fire");
  }
  
  // Checks the putBinding method under normal conditions
  public void testPutBinding_common() {
    checkPutBinding(KeyStroke.getKeyStroke('D', 0), "Dash", 0); // Start
    checkPutBinding(KeyStroke.getKeyStroke('J', 0), "Jump", 5); // Middle
    checkPutBinding(KeyStroke.getKeyStroke('C', 0), "Climb", this.testPanel.getBindingCount()); // End
  }
  
  /*
   * Checks variant of the putBinding method with the method signature of putBinding(KeyStroke,
   * String) which should add to end
   */
  public void testPutBinding_noIndex() {
    this.testPanel.setCheckAssertions(true);
    int index = this.testPanel.getBindingCount();
    KeyStroke shortcut = KeyStroke.getKeyStroke('F', 0);
    String action = "Fall";
    this.testPanel
        .assertUpdate(new BindingTestPanel.UpdateAssertion(index, shortcut, action, true));
    assertTrue(this.testPanel.putBinding(shortcut, action));
    assertFalse(this.testPanel.isUpdateAssertionSet());
  }
  
  // Checks the putBinding method when adding a duplicate shortcut
  public void testPutBinding_duplicate() {
    int index = 4;
    String action = "Detonate";
    BindingEntry duplicateEntry = this.testPanel.getBindings().get(index);
    checkPutBinding(duplicateEntry.getShortcut(), action, index + 3);
    assertEquals(action, duplicateEntry.getAction());
    assertEquals(duplicateEntry, this.testPanel.getBindings().get(index));
  }
  
  // Checks the putBinding method when adding multiple disabled entries
  public void testPutBinding_disabled() {
    checkPutBinding(BindingEntry.DISABLED, "Pop", 6);
    checkPutBinding(BindingEntry.DISABLED, "Plop", 4);
    checkPutBinding(BindingEntry.DISABLED, "Fizz", 4);
  }
  
  // Checks the putBinding method when adding to indices that are out of bounds
  public void testPutBinding_outOfBounds() {
    checkPutBinding(KeyStroke.getKeyStroke('Y', 0), "Yawn", -1);
    checkPutBinding(KeyStroke.getKeyStroke('Y', 0), "Yawn", this.testPanel.getBindingCount() + 1);
  }
  
  /*-
   * Checks both putBiniding methods that accept an index argument, generating assertions and
   * checking output. Adjusts assertions accordingly for:
   * Index out of bounds
   * Adding a duplicate entry
   */
  private void checkPutBinding(KeyStroke shortcut, String action, int index) {
    // TODO: Currently doesn't handle case of adding disabled entries with duplicate actions.
    assert !this.testPanel.isUpdateAssertionSet() : "Leftover assertions may confuse test results.";
    boolean previousChecking = this.testPanel.isAssertionsChecked();
    this.testPanel.setCheckAssertions(true);
    int startingSize = this.testPanel.getBindingCount();
    BindingEntry newEntry = new BindingEntry(shortcut, action);
    
    if (index < 0 || index > startingSize) {
      // Test for adding an entry that's out of bounds
      try {
        this.testPanel.putBinding(shortcut, action, index);
        fail();
      } catch (IndexOutOfBoundsException exc) {
        // Successfully threw exception.
      }
      
      try {
        this.testPanel.putBinding(newEntry, index);
        fail();
      } catch (IndexOutOfBoundsException exc) {
        // Successfully threw exception.
      }
      
      assertEquals(startingSize, this.testPanel.getBindingCount());
    } else if (shortcut != BindingEntry.DISABLED && this.testPanel.contains(shortcut)) {
      // Test for adding a duplicate entry
      assertFalse(this.testPanel.putBinding(shortcut, action, index));
      assertFalse(this.testPanel.putBinding(newEntry, index));
      assertEquals(startingSize, this.testPanel.getBindingCount());
    } else {
      // Checks via the putBinding(KeyStroke, String, index) method
      BindingTestPanel.UpdateAssertion updateAssertion =
          new BindingTestPanel.UpdateAssertion(index, shortcut, action, true);
      this.testPanel.assertUpdate(updateAssertion);
      assertShift(true, index);
      assertTrue(this.testPanel.putBinding(shortcut, action, index));
      assertEquals(startingSize + 1, this.testPanel.getBindingCount());
      assertFalse(this.testPanel.isUpdateAssertionSet());
      
      // Removes just added entry
      this.testPanel.setCheckAssertions(false);
      this.testPanel.removeBinding(index);
      this.testPanel.setCheckAssertions(true);
      
      // Checks via the putBinding(BindingEntry, index) method
      this.testPanel.assertUpdate(updateAssertion);
      assertShift(true, index);
      assertTrue(this.testPanel.putBinding(newEntry, index));
      assertEquals(startingSize + 1, this.testPanel.getBindingCount());
    }
    
    assertFalse(this.testPanel.isUpdateAssertionSet());
    this.testPanel.setCheckAssertions(previousChecking); // Revert setting
  }
  
  // Checks the removeBinding method under normal conditions
  public void testRemoveBinding_common() {
    checkRemoval(this.testPanel.getBindings().get(0)); // Start
    checkRemoval(this.testPanel.getBindings().get(5)); // Middle
    checkRemoval(this.testPanel.getBindings().get(this.testPanel.getBindingCount() - 1)); // End
  }
  
  // Checks the removeBinding method when removing entries that no longer exist
  public void testRemoveBinding_missing() {
    BindingEntry removalEntry = this.testPanel.getBindings().get(3);
    checkRemoval(removalEntry);
    checkRemoval(removalEntry);
  }
  
  // Checks the removeBinding method when removing from indices that are out of bounds
  public void testRemoveBinding_outOfBounds() {
    int initialSize = this.testPanel.getBindingCount();
    try {
      this.testPanel.removeBinding(-1);
      fail();
    } catch (IndexOutOfBoundsException exc) {
      // Successfully threw exception.
    }
    
    try {
      this.testPanel.removeBinding(this.testPanel.getBindingCount());
      fail();
    } catch (IndexOutOfBoundsException exc) {
      // Successfully threw exception.
    }
    assertEquals(initialSize, this.testPanel.getBindingCount());
  }
  
  // Checks the removeBinding method when attempting to remove all entries
  public void testRemoveBinding_all() {
    for (BindingEntry entry : this.testPanel.getBindings()) {
      checkRemoval(entry);
    }
    assertEquals(0, this.testPanel.getBindingCount());
  }
  
  /*- 
   * Checks both removeBiniding methods, generating assertions and checking output. Assertions are
   * adjusted accordingly for:
   * Entry not contained in collection
   */
  private void checkRemoval(BindingEntry entry) {
    assert !this.testPanel.isUpdateAssertionSet() : "Leftover assertions may confuse test results.";
    boolean previousChecking = this.testPanel.isAssertionsChecked();
    this.testPanel.setCheckAssertions(true);
    ArrayList <BindingEntry> bindings = this.testPanel.getBindings();
    int startingSize = this.testPanel.getBindingCount();
    int removalIndex = bindings.indexOf(entry);
    
    if (removalIndex != -1) {
      // Checks via the removeBinding(Entry) method
      assertShift(false, removalIndex + 1);
      assertTrue(this.testPanel.removeBinding(entry));
      assertFalse(this.testPanel.isUpdateAssertionSet());
      assertEquals(startingSize - 1, this.testPanel.getBindingCount());
      
      // Re-adds removed entry
      this.testPanel.setCheckAssertions(false);
      this.testPanel.putBinding(entry, removalIndex);
      this.testPanel.setCheckAssertions(true);
      
      // Checks via the removeBinding(int) method
      assertShift(false, removalIndex + 1);
      assertEquals(entry, this.testPanel.removeBinding(removalIndex));
      assertEquals(startingSize - 1, this.testPanel.getBindingCount());
    } else {
      assertFalse(this.testPanel.removeBinding(entry));
      assertEquals(startingSize, this.testPanel.getBindingCount());
    }
    
    assertFalse(this.testPanel.isUpdateAssertionSet());
    this.testPanel.setCheckAssertions(previousChecking); // Revert setting
  }
  
  // Checks contains method
  public void testContains() {
    this.testPanel.setCheckAssertions(false);
    
    // Confirms that the method recognizes all test values, but fails to find them in an empty panel
    BindingTestPanel empty = new BindingTestPanel();
    empty.setCheckAssertions(false);
    for (BindingEntry entry : this.testPanel.getBindings()) {
      if (entry.getShortcut() == BindingEntry.DISABLED) continue;
      assertTrue(this.testPanel.contains(entry.getShortcut()));
      assertFalse(empty.contains(entry.getShortcut()));
    }
    
    // Checks disabled shortcuts
    assertFalse(this.testPanel.contains(BindingEntry.DISABLED));
    assertFalse(empty.contains(BindingEntry.DISABLED));
    
    this.testPanel.putBinding(BindingEntry.DISABLED, "Foo");
    empty.putBinding(BindingEntry.DISABLED, "Foo");
    
    assertTrue(this.testPanel.contains(BindingEntry.DISABLED));
    assertTrue(empty.contains(BindingEntry.DISABLED));
  }
  
  // Checks collection size verses the getBindingCount method
  public void testGetBindingCount() {
    ArrayList <BindingEntry> entries = this.testPanel.getBindings();
    int size = entries.size();
    this.testPanel.setCheckAssertions(false);
    assertEquals(size, this.testPanel.getBindingCount());
    
    // Checks when removing entries
    for (BindingEntry entry : entries) {
      this.testPanel.removeBinding(entry);
      assertEquals(--size, this.testPanel.getBindingCount());
    }
    
    // Checks when performing removals that shouldn't change the size
    for (BindingEntry entry : entries) {
      this.testPanel.removeBinding(entry);
      assertEquals(0, this.testPanel.getBindingCount());
    }
    
    // Checks when re-adding entries
    for (BindingEntry entry : entries) {
      this.testPanel.putBinding(entry, 0);
      assertEquals(++size, this.testPanel.getBindingCount());
    }
    
    // Checks when performing adds that shouldn't change the size
    for (BindingEntry entry : entries) {
      this.testPanel.putBinding(entry, 0);
      assertEquals(size, this.testPanel.getBindingCount());
    }
  }
  
  // Checks that the mapping returned by a Binding panel equates whatever it was constructed with
  public void testRegression() {
    LinkedHashMap <KeyStroke, String> bindings = new LinkedHashMap <KeyStroke, String>();
    bindings.put(KeyStroke.getKeyStroke('B', 0), "Back");
    bindings.put(KeyStroke.getKeyStroke('F', 0), "Forward");
    bindings.put(KeyStroke.getKeyStroke('R', KeyEvent.CTRL_MASK), "Redo");
    bindings.put(KeyStroke.getKeyStroke('Z', KeyEvent.CTRL_MASK), "Undo");
    bindings.put(KeyStroke.getKeyStroke('T', KeyEvent.CTRL_MASK), "New Tab");
    bindings.put(KeyStroke.getKeyStroke(KeyEvent.VK_TAB, KeyEvent.CTRL_MASK), "Next Tab");
    
    // Constructs panel with mapped bindings
    BindingTestPanel panel = new BindingTestPanel();
    panel.setCheckAssertions(false);
    for (KeyStroke shortcut : bindings.keySet()) {
      panel.putBinding(shortcut, bindings.get(shortcut));
    }
    
    // Retrieves bindings from panel, checking that they equal initial values
    LinkedHashMap <KeyStroke, String> panelBindings = new LinkedHashMap <KeyStroke, String>();
    for (BindingEntry entry : panel.getBindings()) {
      KeyStroke shortcut = entry.getShortcut();
      String action = entry.getAction();
      panelBindings.put(shortcut, action);
    }
    assertEquals(bindings, panelBindings);
  }
  
  // Clicks in the middle of each field to check for basic functionality.
  public void testClick_common() {
    this.testPanel.setCheckAssertions(true);
    BindingEntry sample = this.testPanel.getBindings().get(0);
    int entryHeight = sample.getPreferredSize().height;
    int bottom = 0; // Bottom coordinate of the last entry checked
    
    // Header for textual component of visual demo
    if (LAYOUT_DEMO_DURATION > 0) {
      System.out.println();
      System.out.println("Common test for clicking:");
      
      DEMO_FRAME.add(this.testPanel);
      DEMO_FRAME.pack();
      DEMO_FRAME.setVisible(true);
    }
    
    for (BindingEntry entry : this.testPanel.getBindings()) {
      // Note: this depends on the fields being ordered left to right
      int right = 0; // Right coordinate of last field checked
      for (BindingEntry.Field field : BindingEntry.Field.values()) {
        int width = entry.getField(field).getPreferredSize().width;
        int clickX = right + width / 2;
        int clickY = bottom + (entryHeight / 2);
        
        // Output for visual demo
        if (LAYOUT_DEMO_DURATION > 0) {
          String fieldName = field.toString().toLowerCase();
          System.out.println("\tClick in " + fieldName + " field at " + clickX + "/" + clickY);
          drawClick(clickX, clickY, this.testPanel);
        }
        
        BindingTestPanel.ClickAssertion assertion =
            new BindingTestPanel.ClickAssertion(entry, field);
        this.testPanel.assertClick(assertion);
        
        this.testPanel.doClick(clickX, clickY);
        right += width;
        assertFalse(this.testPanel.isClickAssertionSet());
      }
      
      bottom += entryHeight;
    }
    
    // Ends visual demo for this test
    if (LAYOUT_DEMO_DURATION > 0) {
      DEMO_FRAME.remove(this.testPanel);
      DEMO_FRAME.dispose();
    }
  }
  
  // Checks for appropriate shift when fields and entries may not be visible.
  public void testClick_variantVisibility() {
    BindingTestPanel bindingTestPanel = new BindingTestPanel();
    bindingTestPanel.setCheckAssertions(false);
    bindingTestPanel.putBinding('S', 0, "Up, down, over, through,");
    bindingTestPanel.putBinding('E', 0, "back around, the joke's on you.");
    bindingTestPanel.setCheckAssertions(true);
    
    // Checks all permutations of when fields are and aren't visible
    BindingEntry.Field[] fields = BindingEntry.Field.values();
    int combinations = (int) Math.pow(2, fields.length);
    BindingEntry testEntry = bindingTestPanel.getBindings().get(0); // Entry to test
    
    // Header for textual component of visual demo
    if (LAYOUT_DEMO_DURATION > 0) {
      System.out.println();
      System.out.println("Variant visibility test (invisible fields marked by 'X'):");
      
      DEMO_FRAME.add(bindingTestPanel);
      DEMO_FRAME.pack();
      DEMO_FRAME.setVisible(true);
    }
    
    for (int permutation = 0; permutation < combinations; ++permutation) {
      // Represents each fields visibility status as a bit in the permutation
      boolean[] visibility = new boolean[fields.length]; // Parallel array with visibility status
      for (int index = 0; index < visibility.length; ++index) {
        JLabel fieldLabel = testEntry.getField(fields[index]);
        boolean isVisible = (permutation & (int) Math.pow(2, index)) == 0;
        fieldLabel.setVisible(isVisible);
        visibility[index] = isVisible;
      }
      
      // Output for visual demo
      if (LAYOUT_DEMO_DURATION > 0) {
        // Prints field visibility
        System.out.print("\t");
        for (int i = 0; i < fields.length; ++i) {
          if (visibility[i]) {
            System.out.print(fields[i].toString().substring(0, 1));
          } else {
            System.out.print("X");
          }
        }
        System.out.println();
        
        bindingTestPanel.repaint();
        DEMO_FRAME.pack();
        
        try {
          Thread.sleep(LAYOUT_DEMO_DURATION);
        } catch (InterruptedException exc) {
          exc.printStackTrace();
        }
      }
      
      // Attempts to click on the left side of each field
      int right = 0; // Right coordinate of last visible checked field
      for (int index = 0; index < fields.length; ++index) {
        BindingEntry.Field testField = fields[index]; // Field being checked
        int widthShift = 0; // Width of visible field (for shifting right coordinate)
        BindingTestPanel.ClickAssertion assertion;
        
        if (!visibility[0] && !visibility[1] && !visibility[2]) {
          // If no fields are visible then click falls on the next line
          BindingEntry nextEntry = bindingTestPanel.getBindings().get(1);
          assertion = new BindingTestPanel.ClickAssertion(nextEntry, testField);
          widthShift = testEntry.getField(testField).getPreferredSize().width;
        } else if (visibility[index]) {
          // Field is visible, so assert clicking finds it
          assertion = new BindingTestPanel.ClickAssertion(testEntry, testField);
          widthShift = testEntry.getField(testField).getPreferredSize().width;
        } else {
          // Gets the next visible field (null if none exists)
          BindingEntry.Field nextField = null;
          for (int i = index + 1; i < fields.length; ++i) {
            if (visibility[i]) {
              nextField = fields[i];
              break;
            }
          }
          
          assertion = new BindingTestPanel.ClickAssertion(testEntry, nextField);
        }
        
        // Indicates clicked position
        if (LAYOUT_DEMO_DURATION > 0) {
          drawClick(right + 1, 5, bindingTestPanel);
        }
        
        bindingTestPanel.assertClick(assertion);
        bindingTestPanel.doClick(right + 1, 5);
        assertFalse(bindingTestPanel.isClickAssertionSet());
        right += widthShift;
      }
    }
    
    // Ends visual demo for this test
    if (LAYOUT_DEMO_DURATION > 0) {
      DEMO_FRAME.remove(bindingTestPanel);
      DEMO_FRAME.dispose();
    }
  }
  
  /*
   * Checks for appropriate response when additional components are added without changing the
   * layout.
   */
  public void testClick_extraComponents() {
    /*
     * Checks all permutations where a labels are inserted into the entry. At most a single addition
     * is made at any particular location (between components or at the ends).
     */
    BindingTestPanel bindingTestPanel = new BindingTestPanel();
    BindingEntry.Field[] fields = BindingEntry.Field.values();
    int combinations = (int) Math.pow(2, fields.length + 1);
    
    // Header for textual component of visual demo
    if (LAYOUT_DEMO_DURATION > 0) {
      System.out.println();
      System.out.println("Extra components test (labels represented by '*'):");
      
      DEMO_FRAME.add(bindingTestPanel);
      DEMO_FRAME.setVisible(true);
    }
    
    int insertWidth = 100;
    int insertHeight = 40; // Note: test will fail if this isn't greater than all field heights
    for (int permutation = 0; permutation < combinations; ++permutation) {
      // Reverts panel to have a single binding without extra labels
      bindingTestPanel.setCheckAssertions(false);
      if (bindingTestPanel.getBindingCount() > 0) bindingTestPanel.removeBinding(0);
      BindingEntry testEntry = new BindingEntry(KeyStroke.getKeyStroke('H', 0), "Ho, hum");
      bindingTestPanel.putBinding(testEntry, 0);
      bindingTestPanel.setCheckAssertions(true);
      
      // Sets background color of fields to be distinguishable from background
      for (BindingEntry.Field field : fields) {
        testEntry.getField(field).setOpaque(true);
        testEntry.getField(field).setBackground(Color.BLUE);
      }
      
      // Represents each inserting position as a bit in the permutation
      boolean[] insertPositions = new boolean[fields.length + 1]; // Label inserting positions
      int insertCount = 0; // Number of labels inserted so far
      for (int index = 0; index < insertPositions.length; ++index) {
        boolean addInsert = (permutation & (int) Math.pow(2, index)) != 0;
        insertPositions[index] = addInsert;
        
        if (addInsert) {
          JLabel insert = new JLabel("<INSERT>");
          insert.setOpaque(true);
          insert.setBackground(Color.WHITE);
          insert.setPreferredSize(new Dimension(insertWidth, insertHeight));
          testEntry.add(insert, insertCount + index);
          ++insertCount;
        }
      }
      
      /*
       * Skips test where no inserts are included (later tests assume the entry's preferred height
       * is equal to the insert height... besides, this is covered in the testClick_common test)
       */
      if (insertCount == 0) continue;
      
      // Output for visual demo
      if (LAYOUT_DEMO_DURATION > 0) {
        // Prints layout
        System.out.print("\t");
        for (int index = 0; index < insertPositions.length; ++index) {
          if (insertPositions[index]) System.out.print("*");
          
          if (index < fields.length) {
            System.out.print(fields[index].toString().substring(0, 1));
          }
        }
        System.out.println();
        
        bindingTestPanel.repaint();
        DEMO_FRAME.pack();
        
        try {
          Thread.sleep(LAYOUT_DEMO_DURATION);
        } catch (InterruptedException exc) {
          exc.printStackTrace();
        }
      }
      
      // Attempts to click on the left side of each component (both fields and inserted label)
      int right = 0; // Right coordinate of last visible checked field
      for (int index = 0; index < insertPositions.length + fields.length; ++index) {
        BindingEntry.Field field; // Field occupying this width (null if insert)
        int widthShift = 0; // Width of component being checked (for shifting right coordinate)
        
        if (index % 2 == 0) {
          // Index belongs to an insert- skip if none's there, otherwise assert no field is clicked
          if (!insertPositions[index / 2]) continue;
          field = null;
          widthShift = insertWidth;
        } else {
          // Index belongs to field
          field = fields[(index - 1) / 2];
          widthShift = testEntry.getField(field).getPreferredSize().width;
        }
        
        /*
         * Clicks on the top, middle, and bottom of the entry. If a field only the middle click
         * should detect it (this is hard coded as index one).
         */
        int[] clickHeights = new int[3];
        clickHeights[0] = 1;
        clickHeights[1] = insertHeight / 2;
        clickHeights[2] = insertHeight - 1;
        
        for (int i = 0; i < clickHeights.length; ++i) {
          // Indicates clicked position
          if (LAYOUT_DEMO_DURATION > 0) {
            drawClick(right + 1, clickHeights[i], bindingTestPanel);
          }
          
          BindingEntry.Field clickedField = i == 1 ? field : null;
          BindingTestPanel.ClickAssertion assertion =
              new BindingTestPanel.ClickAssertion(testEntry, clickedField);
          bindingTestPanel.assertClick(assertion);
          bindingTestPanel.doClick(right + 1, clickHeights[i]);
          assertFalse(bindingTestPanel.isClickAssertionSet());
        }
        right += widthShift;
      }
    }
    
    // Ends visual demo for this test
    if (LAYOUT_DEMO_DURATION > 0) {
      DEMO_FRAME.remove(bindingTestPanel);
      DEMO_FRAME.dispose();
    }
  }
  
  // Draws a dot in the panel to indicate where clicks are made
  private void drawClick(int x, int y, JPanel panel) {
    // TODO: Currently draws via a hack- this functionality should be in the BindingTestPanel.
    if (CLICK_DURATION < 1) return;
    int dotRadius = 3;
    panel.setIgnoreRepaint(true);
    Graphics2D g2 = (Graphics2D) panel.getGraphics();
    g2.setColor(Color.RED);
    g2.fillOval(x - dotRadius, y - dotRadius, dotRadius * 2, dotRadius * 2);
    
    try {
      Thread.sleep(CLICK_DURATION);
    } catch (InterruptedException exc) {
      exc.printStackTrace();
    }
  }
  
  // Asserts an update in the test display for entries from an index to the end.
  private void assertShift(boolean isForward, int startingIndex) {
    for (int i = 0; i < this.testPanel.getBindings().size(); ++i) {
      if (i < startingIndex) continue;
      BindingEntry entry = this.testPanel.getBindings().get(i);
      KeyStroke shortcut = entry.getShortcut();
      String action = entry.getAction();
      
      int newIndex = isForward ? i + 1 : i - 1;
      BindingTestPanel.UpdateAssertion updateAssertion =
          new BindingTestPanel.UpdateAssertion(newIndex, shortcut, action, false);
      this.testPanel.assertUpdate(updateAssertion);
    }
  }
}
