package test.chooser;

import java.awt.event.MouseEvent;
import java.util.LinkedList;

import javax.swing.KeyStroke;

import chooser.BindingEntry;
import chooser.BindingPanel;

import junit.framework.TestCase;

/**
 * Testable implementation of the BindingPanel class allowing assertions of calls to the onUpdate
 * and onClick methods. If there's a mismatch or unasserted firing then this will prompt a failure
 * in the current test. Asserted values are removed when matched to a call of their method.
 * @author Damian Johnson (atagar1@gmail.com)
 * @version September 1, 2007
 */
class BindingTestPanel extends BindingPanel {
  private static final long serialVersionUID = 0;
  private boolean checkAssertions = true;
  private LinkedList <UpdateAssertion> updateAssertions = new LinkedList <UpdateAssertion>();
  private LinkedList <ClickAssertion> clickAssertions = new LinkedList <ClickAssertion>();
  
  // Tests that update matches what was expected
  protected void onUpdate(int index, BindingEntry entry, boolean isNew) {
    if (!this.checkAssertions) return;
    
    if (this.updateAssertions.isEmpty()) {
      TestCase.fail();
    } else {
      UpdateAssertion call = new UpdateAssertion(index, entry, isNew);
      boolean isMatch = this.updateAssertions.getFirst().equals(call);
      TestCase.assertTrue(isMatch);
      if (isMatch) this.updateAssertions.remove(call);
    }
  }
  
  // Tests that click matches what was expected
  protected void onClick(MouseEvent event, BindingEntry entry, BindingEntry.Field field) {
    if (!this.checkAssertions) return;
    
    if (this.clickAssertions.isEmpty()) {
      TestCase.fail();
    } else {
      ClickAssertion call = new ClickAssertion(entry, field);
      boolean isMatch = this.clickAssertions.getFirst().equals(call);
      TestCase.assertTrue(isMatch);
      if (isMatch) this.clickAssertions.remove(call);
    }
  }
  
  /**
   * Sets if the onUpdate and onClick method calls should be checked against asserted values. If
   * disabled then this class will be unable to throw failures and method calls won't clear asserted
   * values.
   * @param check if true onUpdate and onClick method calls will be checked against asserted values,
   *          otherwise they will be ignored
   */
  protected void setCheckAssertions(boolean check) {
    this.checkAssertions = check;
  }
  
  /**
   * Returns if the onUpdate and onClick method calls are checked against asserted values.
   * @return true if assertions are checked, false otherwise
   */
  protected boolean isAssertionsChecked() {
    return this.checkAssertions;
  }
  
  /**
   * Enqueues set of values to check against future call of the onUpdate method.
   * @param assertion expected values of a later call to the onUpdate method
   */
  protected void assertUpdate(UpdateAssertion assertion) {
    this.updateAssertions.add(assertion);
  }
  
  /**
   * Enqueues set of values to check against future call of the onClick method.
   * @param assertion expected values of a later call to the onClick method
   */
  protected void assertClick(ClickAssertion assertion) {
    this.clickAssertions.add(assertion);
  }
  
  /**
   * Clears asserted values for the onUpdate and onClick methods.
   */
  protected void clearAssertions() {
    this.updateAssertions.clear();
    this.clickAssertions.clear();
  }
  
  /**
   * Returns if awaiting an unfulfilled assertion of the onUpdate method.
   * @return true if currently awaiting a firing of the onUpdate method, false otherwise
   */
  protected boolean isUpdateAssertionSet() {
    return !this.updateAssertions.isEmpty();
  }
  
  /**
   * Returns if awaiting an unfulfilled assertion of the onClick method.
   * @return true if currently awaiting a firing of the onClick method, false otherwise
   */
  protected boolean isClickAssertionSet() {
    return !this.clickAssertions.isEmpty();
  }
  
  /**
   * Emulates a mouse click within the display to check the onClick method.
   * @param x horizontal pixel coordinate of click
   * @param y vertical pixel coordinate of click
   */
  protected void doClick(int x, int y) {
    // Allows content to be detected properly (similar functionality to the pack method)
    setSize(getPreferredSize());
    validateTree();
    
    long time = System.currentTimeMillis();
    int mod = MouseEvent.NOBUTTON;
    int clicks = 1;
    MouseEvent event = new MouseEvent(this, 0, time, mod, x, y, clicks, false);
    this.getMouseListeners()[0].mousePressed(event);
  }
  
  /**
   * Container for an expected firing of the BindingPanel's onUpdate method.
   */
  static class UpdateAssertion {
    private int index;
    private KeyStroke shortcut;
    private String action;
    private boolean isNew;
    
    protected UpdateAssertion(int index, BindingEntry entry, boolean isNew) {
      this(index, entry.getShortcut(), entry.getAction(), isNew);
    }
    
    protected UpdateAssertion(int index, KeyStroke shortcut, String action, boolean isNew) {
      this.index = index;
      this.shortcut = shortcut;
      this.action = action;
      this.isNew = isNew;
    }
    
    @Override
    public boolean equals(Object obj) {
      if (obj == this) return true;
      else if (!(obj instanceof UpdateAssertion)) return false;
      
      UpdateAssertion assertion = (UpdateAssertion) obj;
      boolean equals = true;
      equals &= this.index == assertion.index;
      equals &= this.isNew == assertion.isNew;
      
      if (this.shortcut == null) equals &= assertion.shortcut == null;
      else equals &= this.shortcut.equals(assertion.shortcut);
      
      if (this.action == null) equals &= assertion.action == null;
      else equals &= this.action.equals(assertion.action);
      
      return equals;
    }
    
    @Override
    public int hashCode() {
      int hash = 17;
      hash = 37 * hash + this.index;
      hash = 37 * hash + (this.shortcut == null ? 0 : this.shortcut.hashCode());
      hash = 37 * hash + (this.action == null ? 0 : this.action.hashCode());
      hash = 37 * hash + (this.isNew ? 0 : 1);
      return hash;
    }
  }
  
  /**
   * Container for an expected firing of the BindingPanel's onClick method.
   */
  static class ClickAssertion {
    private KeyStroke shortcut;
    private String action;
    private BindingEntry.Field field;
    
    protected ClickAssertion(BindingEntry entry, BindingEntry.Field field) {
      this(entry.getShortcut(), entry.getAction(), field);
    }
    
    protected ClickAssertion(KeyStroke shortcut, String action, BindingEntry.Field field) {
      this.shortcut = shortcut;
      this.action = action;
      this.field = field;
    }
    
    @Override
    public boolean equals(Object obj) {
      if (obj == this) return true;
      else if (!(obj instanceof ClickAssertion)) return false;
      
      ClickAssertion assertion = (ClickAssertion) obj;
      boolean equals = true;
      
      if (this.shortcut == null) equals &= assertion.shortcut == null;
      else equals &= this.shortcut.equals(assertion.shortcut);
      
      if (this.action == null) equals &= assertion.action == null;
      else equals &= this.action.equals(assertion.action);
      
      equals &= this.field == assertion.field; // Either may be null
      return equals;
    }
    
    @Override
    public int hashCode() {
      int hash = 17;
      hash = 37 * hash + (this.shortcut == null ? 0 : this.shortcut.hashCode());
      hash = 37 * hash + (this.action == null ? 0 : this.action.hashCode());
      hash = 37 * hash + (this.field == null ? 0 : this.field.hashCode());
      return hash;
    }
  }
}
