package test.chooser;

import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.io.*;
import java.text.ParseException;
import java.util.LinkedHashMap;
import java.util.Scanner;

import javax.swing.KeyStroke;

import chooser.Persistence;

import junit.framework.TestCase;

/**
 * Tests keybinding persistence (both loading and saving) with a variety of content.
 * @author Damian Johnson (atagar1@gmail.com)
 * @version September 22, 2007
 */
public class PersistenceTest extends TestCase {
  private static final String TEST_FILENAME = "persistenceTestOutput";
  
  public static void main(String args[]) {
    junit.textui.TestRunner.run(PersistenceTest.class);
  }
  
  /**
   * Performs basic verification tests on the following types of content:<br>
   * Empty mapping<br>
   * General content<br>
   * Null content
   */
  public void testGeneral() {
    // Test with an empty mapping
    LinkedHashMap <KeyStroke, String> testContent = new LinkedHashMap <KeyStroke, String> ();
    checkContent(testContent);
    
    // Test with general content
    testContent.put(KeyStroke.getKeyStroke(KeyEvent.VK_UP, 0), "Move Up");
    testContent.put(KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT, 0), "Move Right");
    testContent.put(KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, 0), "Move Down");
    testContent.put(KeyStroke.getKeyStroke(KeyEvent.VK_LEFT, 0), "Move Left");
    testContent.put(KeyStroke.getKeyStroke(KeyEvent.VK_CONTROL, 0), "Next Weapon");
    testContent.put(KeyStroke.getKeyStroke(KeyEvent.VK_SPACE, 0), "Fire");
    testContent.put(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, InputEvent.CTRL_MASK
        | InputEvent.ALT_MASK), "Kill Process");
    checkContent(testContent);
    
    // Test with null content (should fail)
    try {
      testContent.put(null, "Lazy key");
      checkContent(testContent);
      fail();
    } catch (UnsupportedOperationException exc) {
      // Test passes
    }
  }
  
  /**
   * Attempts to save and load bindings with each type of persistence, performing the following checks:<br>
   * Saved content matches the getOutput method output.<br>
   * Loading is equivalent to the original content.
   * @param bindings set of keybindings to be tested
   */
  private void checkContent(LinkedHashMap <KeyStroke, String> bindings) {
    for (Persistence persistenceType : Persistence.values()) {
      // Saves bindings to a unique file in the operating system's temporary directory
      File testFile = null;
      try {
        testFile = File.createTempFile(TEST_FILENAME, "");
        testFile.deleteOnExit();
        persistenceType.save(testFile.getAbsolutePath(), bindings);
      } catch (IOException exc) {
        fail("Unable to save content to a temporary file");
      }
      
      // Checks the getOutput method provides saved content
      try {
        // Needs to scan a FileInputStream rather than File to interpret unparsable characters
        FileInputStream savedStream = new FileInputStream(testFile);
        Scanner scanner = new Scanner(savedStream);
        StringBuilder savedOutput = new StringBuilder();
        
        if (scanner.hasNextLine()) savedOutput.append(scanner.nextLine());
        while (scanner.hasNextLine()) {
          savedOutput.append("\n");
          savedOutput.append(scanner.nextLine());
        }
        
        assertEquals(persistenceType.getOutput(bindings), savedOutput.toString());
        savedStream.close();
        scanner.close();
      } catch (FileNotFoundException exc) {
        fail("Unable to load test content");
      } catch (IOException exc) {
        // Failed to release input stream, but finalizer will handle this
      }
      
      // Checks that loading the saved content provides the original bindings
      try {
        LinkedHashMap <KeyStroke, String> loadedBindings = persistenceType.load(testFile.getAbsolutePath());
        assertEquals(bindings, loadedBindings);
      } catch (IOException exc) {
        fail("Unable to load test content");
      } catch (ParseException exc) {
        fail("Load method is unable to parse saved content: " + exc.getMessage());
      }
    }
  }
}
